/* 

Copyright (c) 2012, Ludovic Jacomme
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

  Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/
#include <stdio.h>
#include <string.h>

int main( int argc, char ** argv )
{
  char buffer[512];
  int  base = 0;
  int  sound = 0;

  int         repeat         = 0;
  const char* in_filename    = "wav2cvdb.as";
  const char* out_filename_s = "sounds_data.s";
  const char* out_filename_c = "sounds.c";

  int i;
  int f    = 0;
  int help = 0;

  for (i = 1; i < argc; ++i) {
    if ( ! strcmp( argv[i], "-repeat")) repeat = 1;
    else 
    if ( ! strcmp( argv[i], "-help")) help = 1;
    else
    if (f == 0) { in_filename = argv[i]; f++; }
    else
    if (f == 1) { out_filename_s = argv[i]; f++; }
    else
    if (f == 2) { out_filename_c = argv[i]; f++; }
    else help = 1;
  }

  if (help) {
    printf("syntax cvdb2s [-repeat] [<input> [<output.s> <output.c>]]\n");
    return 1;
  }

  FILE* fd_in = fopen( in_filename, "r" );
  if (! fd_in) {
    printf("can't open input file '%s'\n", in_filename );
    return 1;
  }
  FILE* fd_out_s = fopen( out_filename_s, "w" );
  if (! fd_out_s) {
    printf("can't open output file '%s'\n", out_filename_s );
    return 1;
  }
  FILE* fd_out_c = fopen( out_filename_c, "w" );
  if (! fd_out_c) {
    printf("can't open output file '%s'\n", out_filename_c );
    return 1;
  }

  printf( "Processing input file '%s'\n", in_filename );
  // On cherche le nombre de tableaux de son
  sound = 0;
  while (fgets( buffer, sizeof( buffer ), fd_in)) {
    char* scan = strchr( buffer, '\r');
    if (scan) *scan = 0;
    scan = strchr( buffer, '\n');
    if (scan) *scan = 0;
    if (! strncmp( buffer, "_wav2cv_sound", 13)) sound++;
  }
  fseek( fd_in, 0L, SEEK_SET);

  printf( "Generating output file '%s'\n", out_filename_c );

  fprintf( fd_out_c, "#include <coleco.h>\n\n" );
  int s;
  for (s = 0; s < sound; ++s) {
    fprintf( fd_out_c, "extern byte wav2cv_sound%d[];\n", s);
  }
  fprintf( fd_out_c, "const sound_t snd_table[] = {" );
  for (s = 0; s < sound; ++s) {
    fprintf( fd_out_c, "\n{ wav2cv_sound%d,SOUNDAREA%d}", s, s+1 );
    if (s != (sound - 1)) fprintf( fd_out_c, ",");
  }
  fprintf( fd_out_c, "\n};\n");
  fclose( fd_out_c );

  fprintf(fd_out_s, ".module music\n");
  for (s = 0; s < sound; ++s) {
    fprintf(fd_out_s, ".globl _wav2cv_sound%d\n", s);
  }
  fprintf(fd_out_s, ".area _CODE\n\n");

  printf( "Generating output file '%s'\n", out_filename_s );

  while (fgets( buffer, sizeof( buffer ), fd_in)) {
    char* scan = strchr( buffer, '\r');
    if (scan) *scan = 0;
    scan = strchr( buffer, '\n');
    if (scan) *scan = 0;
    
    // Ligne a scanner 
    if (! strncmp( buffer, "        defb    ", 16)) {
      if (! base) continue;

      scan = buffer + 16;
      // On remplace tous les 'h' par des espaces pour commencer
      for (i = 0; scan[i]; i++) {
        if (scan[i] == 'h') scan[i] = ' ';
      }

      // fprintf(stdout, "scan line '%s'\n", scan );

      char* strtok_ptr = NULL;
      char* scan_token = strtok_r( scan, ",", &strtok_ptr);

      int   token_num = 0;
      int   token_val[64];
      while (scan_token) {
        int ret = sscanf( scan_token, "%x", &token_val[token_num] );
        if (ret == 1) token_num++;
        else break;
        scan_token = strtok_r( NULL, ",", &strtok_ptr);
        if (token_num >= 64) break; // should not happen
      }
      int pos = 0;
      while (token_num >= 2) {
        int codop = token_val[pos];
        int step  = 1;
        // Silence
        if (codop == 0x9f) {
          // Analyse la valeur suivante
          step = 2;
          if (token_num < step) break; // error ?
          int silence = base + 0x20;
          silence |= token_val[pos+1] & 0x0f;
          if (pos == 0) fprintf( fd_out_s, ".db 0x%x", silence );
          else          fprintf( fd_out_s, ", 0x%x", silence );

        } else
        // Note simple
        if ((codop >= 0x80)  &&
            (codop <= 0xef)) {
          // Analyse les 4 valeurs suivantes
          step = 4;
          if (token_num < step) break; // error ?
          if (pos == 0) fprintf( fd_out_s, ".db 0x%x", base );
          else          fprintf( fd_out_s, ", 0x%x", base );
          fprintf( fd_out_s, ", 0x%x", token_val[pos+1] );
          int third = ((token_val[pos+2] & 0x0f) << 4) | (codop & 0x0f);
          fprintf( fd_out_s, ", 0x%x", third            );
          fprintf( fd_out_s, ", 0x%x", token_val[pos+3] );

        }
        token_num -= step;
        pos += step;
      }
      fprintf( fd_out_s, "\n");

    } else
    // Ligne de fin
    if (! strncmp( buffer, "        defw    00000h", 22)) {
      if (! base) continue;
       // repeat code END
       fprintf( fd_out_s, ".db 0x%x\n\n", base + 0x18 );
       base = 0;
    }

    // Changement de base
    if (! strcmp( buffer, "_wav2cv_sound0:")) {
      base = 0x40;
      fprintf( fd_out_s, "%s\n", buffer );
    } else 
    if (! strcmp( buffer, "_wav2cv_sound1:")) {
      base = 0x80;
      fprintf( fd_out_s, "%s\n", buffer );
    } else 
    if (! strcmp( buffer, "_wav2cv_sound2:")) {
      base = 0xc0;
      fprintf( fd_out_s, "%s\n", buffer );
    }
  }

  fclose( fd_in );
  fclose( fd_out_s );

  printf("Done.\n");

  return 0;
}
